﻿using System;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Windows.Forms;
using TcpIP_class;
using DataClass;
using System.Threading;
using SDK_SC_Fingerprint;
using System.Windows.Forms.DataVisualization.Charting;
using System.Drawing;
using System.Collections.Generic;
using System.Linq;



namespace SampleFridge
{
    public partial class MainWindow : Form
    {
        private string _ipAddress, _serialNumber;
        private int _portNumber;

        private string MyIp;

        EventWaitHandle scanDone = new EventWaitHandle(false,EventResetMode.ManualReset);

        private TcpIpClient _tcpClient = new TcpIpClient();
        private TcpNotificationServer _notificationServer;

        // User currently loaded (in user management tab).
        private DeviceGrant _currentUser;
        // Locally stores all granted users.
        private DeviceGrant[] _grantedUsers;
        // FingerPrint reader handler.
        private FingerPrintClass _fpHandler = new FingerPrintClass();

        // Timer interval: refreshing the current fridge temperature value.
        private const int REFRESH_TEMPERATURE_TIMER = 2000;
        // Timer interval: refreshing the connection status.
        private const int REFRESH_STATUS_TIMER = 8000;

        // Microsoft Chart Control (.NET 3.5) instance (Temperature tab).
        private Chart _chartTemperature;
        // Const string index of the chart series which will contain the measures.
        private const string SERIES_T_NAME = "TEMPERATURE_SERIES";

        private static readonly Bitmap RED_ICON = SampleFridge.Properties.Resources.red_ball_icon;
        private static readonly Bitmap YELLOW_ICON = SampleFridge.Properties.Resources.yellow_ball_icon;
        private static readonly Bitmap GREEN_ICON = SampleFridge.Properties.Resources.green_ball_icon;


        public MainWindow()
        {
            InitializeComponent();

            //Get My Ip For server notifivation if used.
            string host = Dns.GetHostName();
            MyIp = Dns.GetHostEntry(Dns.GetHostName()).AddressList.FirstOrDefault(ip => ip.AddressFamily == AddressFamily.InterNetwork).ToString();
            //MyIp = "192.168.1.20";
          
            // Define the String format for listBoxUsers items.
            listBoxUsers.Format += delegate(object sender, ListControlConvertEventArgs e)
            {
                UserClassTemplate newUser = ((DeviceGrant)e.Value).user;
                e.Value = String.Format("{0} {1}", newUser.firstName, newUser.lastName);
            };

            // Define the String format for users' grants.
            comboBoxGrant.Format += delegate(object sender, ListControlConvertEventArgs e)
            {
                e.Value = getEnumDesc.GetEnumDescription((Enum)e.Value);
            };

            // Add all available grant types.
            comboBoxGrant.Items.Add(UserGrant.UG_NONE);
            comboBoxGrant.Items.Add(UserGrant.UG_SLAVE);
            comboBoxGrant.Items.Add(UserGrant.UG_MASTER);
            comboBoxGrant.Items.Add(UserGrant.UG_MASTER_AND_SLAVE);
            comboBoxGrant.SelectedItem = UserGrant.UG_NONE;

            // Initialize the Chart object (layed out in a Panel control)
            InitializeChart();
            panelChart.Controls.Add(_chartTemperature);
        }

        /// <summary>
        /// Initialize our Chart instance, if it has'nt been created yet, or just clear it points.
        /// </summary>
        private void InitializeChart()
        {
            if (_chartTemperature != null)
            {
                _chartTemperature.Series[SERIES_T_NAME].Points.Clear();
            }

            _chartTemperature = new Chart();
            _chartTemperature.ChartAreas.Add(new ChartArea());
            _chartTemperature.Dock = DockStyle.Fill;
            _chartTemperature.BackColor = Color.AliceBlue;

            _chartTemperature.Series.Add(SERIES_T_NAME);
            _chartTemperature.Series[SERIES_T_NAME].ChartType = SeriesChartType.Line;
            _chartTemperature.Series[SERIES_T_NAME].YValueType = ChartValueType.Double;
            _chartTemperature.Series[SERIES_T_NAME].IsValueShownAsLabel = false;
            _chartTemperature.Series[SERIES_T_NAME].ToolTip = "#VALX : #VALY";
            _chartTemperature.Series[SERIES_T_NAME].BackGradientStyle = GradientStyle.VerticalCenter;
            _chartTemperature.Series[SERIES_T_NAME].ShadowOffset = 0;
            _chartTemperature.Series[SERIES_T_NAME].BackSecondaryColor = Color.White;
            _chartTemperature.Series[SERIES_T_NAME].ShadowColor = Color.FromArgb(128, 0, 0);
            _chartTemperature.Series[SERIES_T_NAME].Color = Color.DodgerBlue;
            _chartTemperature.Series[SERIES_T_NAME].BorderWidth = 3;
            _chartTemperature.Series[SERIES_T_NAME].Font = new Font("Verdana", 12.0f, FontStyle.Bold);
        }

        /// <summary>
        /// Click event on button "Connect".
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonConnect_Click(object sender, EventArgs e)
        {
            pictureBoxConnectionStatus.Image = RED_ICON;
            buttonConnect.Enabled = false;
            tabControl.Enabled = false;

            // Stops refreshing timers (In case they have been started during a previous connection).
            SetRefreshingTimers(false);

            // Makes serial textfield empty.
            textBoxSerialNumber.Text = String.Empty;

            string inputIp;
            int inputPort;

            // Checks user inputs.
            if (!Tool.TryParseIpPort(out inputIp, out inputPort, textBoxIp.Text, textBoxPort.Text))
            {
                MessageBox.Show("Please enter valid IP address and port number.", "Invalid parameters", MessageBoxButtons.OK, MessageBoxIcon.Information);
                buttonConnect.Enabled = true;
                return;
            }

            InitializeTCPClient(inputIp, inputPort);
        }

        /// <summary>
        /// Try to establish a TCP connection with the device (server).
        /// Save ip address, port number and device serial number.
        /// Get last scan inventory (from device) and display it.
        /// Executed by the ThreadPool in order not to block the UI thread.
        /// </summary>
        /// <param name="inputIp">IP Address (given by User).</param>
        /// <param name="inputPort">Port Number (given by User).</param>
        private void InitializeTCPClient(string inputIp, int inputPort)
        {
            ThreadPool.QueueUserWorkItem(new WaitCallback(delegate
            {
                pictureBoxConnectionStatus.Image = YELLOW_ICON;
                UpdateStatusBar("Trying to connect to " + inputIp + ":" + inputPort + "...");

                // Store information of the devices found on remote host.
                DeviceInfo[] arrayDevice;
                _tcpClient.getDevice(inputIp, inputPort, out arrayDevice);

                // No device found.
                if (arrayDevice == null || arrayDevice.Length == 0)
                {
                    UpdateStatusBar("Connection failed.");
                    MessageBox.Show("No device found on this host.", "Connection failed", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                    // Called on UI Thread: disable the tab control.
                    Invoke((MethodInvoker)delegate { tabControl.Enabled = false; buttonConnect.Enabled = true; });
                    pictureBoxConnectionStatus.Image = RED_ICON;
                    return;
                }

                // Only one device is used on remote hosts like Fridges ones. This is necessary the first of the array.
                string serialNumber = arrayDevice[0].SerialRFID;

                // Check that the device is responding to a basic ping request.
                if (_tcpClient.pingDevice(inputIp, inputPort, serialNumber) != TcpIpClient.RetCode.RC_Succeed)
                {
                    UpdateStatusBar("Device not responding.");
                    MessageBox.Show(String.Format("Device {0} is not responding ({1}:{2})...", serialNumber, inputIp, inputPort), "Communication issues", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    Invoke((MethodInvoker)delegate { tabControl.Enabled = false; buttonConnect.Enabled = true; });
                    pictureBoxConnectionStatus.Image = RED_ICON;
                    return;
                }

                pictureBoxConnectionStatus.Image = GREEN_ICON;

                // Called on UI Thread: Display the serial number in the dedicated control
                // and enable the Tab control.
                Invoke((MethodInvoker)delegate { textBoxSerialNumber.Text = serialNumber; });
                Invoke((MethodInvoker)delegate { tabControl.Enabled = true; buttonConnect.Enabled = true; });

                _ipAddress = inputIp;
                _portNumber = inputPort;
                _serialNumber = serialNumber;

                int lastScanId;

                UpdateStatusBar("Getting last scan inventory...");

                if (TcpIpClient.RetCode.RC_Succeed == _tcpClient.getLastScanID(_ipAddress, _portNumber, _serialNumber, out lastScanId))
                {
                    UpdateStatusBar("Device ready.");
                    UpdateLastInventory(lastScanId);
                }

                else
                {
                    UpdateStatusBar("Unable to get last scan inventory.");
                }

              

                // Instantiate a notifications handler. 
                _notificationServer = new TcpNotificationServer(_portNumber + 1);
                _notificationServer.TcpNotifyEvent += new TcpNotificationServer.TcpNotifyHandlerDelegate(HandleNotificationEvent);
                _notificationServer.StartServer();

                // Ask to the device (server) to setup event notifications
                _tcpClient.SetTcpServerNotificationInfo(_ipAddress, _portNumber, true, MyIp, _portNumber + 1);

                // Starts the refreshing timers.
                SetRefreshingTimers(true);
            }));
        }

        /// <summary>
        /// Enable/disable refreshing timers (users list & current temperature).
        /// Executed on UI thread (timers requierement).
        /// </summary>
        /// <param name="state">True: enable timers. False: stop timers.</param>
        private void SetRefreshingTimers(bool state)
        {
            Invoke((MethodInvoker)delegate
            {
                if (state)
                {
                    timerRefreshCurrentTemperature.Interval = REFRESH_TEMPERATURE_TIMER;
                    timerRefreshCurrentTemperature.Enabled = true;

                    timerRefreshConnectionStatus.Interval = REFRESH_STATUS_TIMER;
                    timerRefreshConnectionStatus.Enabled = true;
                }

                else
                {
                    timerRefreshCurrentTemperature.Enabled = false;
                    timerRefreshConnectionStatus.Enabled = false;
                }
            });
        }

        /// <summary>
        /// Stop event notifications server, if it has been started.
        /// </summary>
        private void StopNotificationServer()
        {
            if (_notificationServer != null)
            {
                _notificationServer.StopSocket();
            }
        }

        /// <summary>
        /// Handle all kind of event notifications sent by the device (Server).
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="arg">Contains event type and information.</param>
        private void HandleNotificationEvent(Object sender, rfidTcpNotArg arg)
        {
            switch (arg.RN_Value)
            {
                case rfidTcpNotArg.ReaderTcpNotify.RN_ScanStarted:
                    UpdateStatusBar("Scan started!");
                    UpdateScanButtonsState(false);
                    break;

                case rfidTcpNotArg.ReaderTcpNotify.RN_ScanCompleted:
                    UpdateStatusBar("Scan completed!");
                    UpdateScanButtonsState(true);
                    UpdateLastInventory(arg.ScanId);
                    scanDone.Set();
                    break;

                case rfidTcpNotArg.ReaderTcpNotify.RN_Door_Opened:
                    UpdateStatusBar("Door opened!");
                    break;

                case rfidTcpNotArg.ReaderTcpNotify.RN_ScanCancelledByHost:
                    UpdateStatusBar("Scan cancelled!");
                    UpdateScanButtonsState(true);
                    break;
                    case rfidTcpNotArg.ReaderTcpNotify.RN_TempEventChanged:
                    case rfidTcpNotArg.ReaderTcpNotify.RN_TempEvent:
                    Invoke((MethodInvoker)delegate { textBoxCurrentTemperature.Text = String.Format("{0:0.0}", arg.LastTempBottle); });
                    UpdateStatusBar("Temperature received :" + String.Format("{0:0.0}", arg.LastTempBottle));
                    break;
                    case rfidTcpNotArg.ReaderTcpNotify.RN_TestNotification:
                    UpdateStatusBar("Test Notification received");
                    break;

            }
        }

        /// <summary>
        /// Called on ScanCompleted event notification.
        /// Reset all data of scan tab and display last inventory (provided by Device).
        /// </summary>
        /// <param name="scanId">Last scan process ID (provided by ScanCompleted event notification).</param>
        private void UpdateLastInventory(int scanId)
        {
            InventoryData lastInventory = new InventoryData();
            TcpIpClient.RetCode response = _tcpClient.requestGetScanFromIdEvent(_ipAddress, _portNumber, _serialNumber, scanId, out lastInventory);

            Invoke((MethodInvoker)delegate
            {
                listBoxTag.Items.Clear();
                textBoxTagAdded.Text = "0";
                textBoxTagRemoved.Text = "0";
                textBoxTagPresent.Text = "0";
                textBoxTagTotal.Text = "0";

                if (lastInventory != null)
                {
                    foreach (string tagUID in lastInventory.listTagAll)
                    {
                        listBoxTag.Items.Add(tagUID);
                    }

                    textBoxTagAdded.Text = lastInventory.nbTagAdded.ToString();
                    textBoxTagRemoved.Text = lastInventory.nbTagRemoved.ToString();
                    textBoxTagPresent.Text = lastInventory.nbTagPresent.ToString();
                    textBoxTagTotal.Text = lastInventory.nbTagAll.ToString();

                    switch (lastInventory.accessType)
                    {
                        case AccessType.AT_BADGEREADER:
                            textBoxLastScanType.Text = "Badge (" + lastInventory.userLastName + " " + lastInventory.userFirstName + ")";
                            break;

                        case AccessType.AT_FINGERPRINT:
                            textBoxLastScanType.Text = "Fingerprint (" + lastInventory.userLastName + " " + lastInventory.userFirstName + ")";
                            break;

                        case AccessType.AT_NONE:
                        default:
                            textBoxLastScanType.Text = "Manual";
                            break;
                    }

                    // Device (server) sends GMT-formatted event date. Let's convert it to local time.
                    lastInventory.eventDate = DateTime.SpecifyKind(lastInventory.eventDate, DateTimeKind.Utc).ToLocalTime();

                    textBoxLastScanTime.Text = lastInventory.eventDate.ToLocalTime().ToString("G");
                }

                else
                {
                    UpdateStatusBar("Unable to get inventory (id: " + scanId + ") from Device.");
                }
            });
        }

        /// <summary>
        /// Update Start/Stop button state according to the event notification sent by the device.
        /// Executed on UI Thread.
        /// </summary>
        /// <param name="state"></param>
        private void UpdateScanButtonsState(bool state)
        {
            Invoke((MethodInvoker)delegate
            {
                buttonStop.Enabled = !state;
                buttonStart.Enabled = state;
                //disable button light - wil be enable after selection
                buttonLight.Enabled = false;
            });
        }


        /// <summary>
        /// Update stauts label in Status bar. 
        /// Executed on UI Thread.
        /// </summary>
        /// <param name="message">New status to be displayed</param>
        private void UpdateStatusBar(string message)
        {
            Invoke((MethodInvoker)delegate { statusLabel.Text = DateTime.Now.ToLongTimeString()  + " >> " + message; });
        }

        /// <summary>
        /// Click event on button "Add User".
        /// Create a new DeviceGrant (granted user) instance and start an optional (= can be cancelled) fingerprint enrollment process.
        /// If the enrollment is cancelled, the user's fingerprint templates are left empty.
        /// Once the user has been added, users list is automatically refreshed.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonAddUser_Click(object sender, EventArgs e)
        {
            if (String.IsNullOrEmpty(textBoxFirstName.Text.Trim()) || String.IsNullOrEmpty(textBoxLastName.Text.Trim()))
            {
                MessageBox.Show("First and last names are mandatory fields.", "Incomplete form", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                return;
            }

            _currentUser = new DeviceGrant();
            _currentUser.user.firstName = textBoxFirstName.Text.Trim();
            _currentUser.user.lastName = textBoxLastName.Text.Trim();
            _currentUser.userGrant = (UserGrant)comboBoxGrant.SelectedItem;
            _currentUser.user.BadgeReaderID = textBoxBadgeId.Text.Trim();

            DialogResult userChoice = MessageBox.Show("Do you want to process fingerprint enrollment?", "Fingerprint Enrollment", MessageBoxButtons.YesNo);
            _currentUser.user.template = (userChoice != DialogResult.Yes) ? null : _fpHandler.EnrollUser(null, _currentUser.user.firstName, _currentUser.user.lastName, null, false);

            _tcpClient.addUserFromTemplate(_ipAddress, _portNumber, _currentUser.user.firstName, _currentUser.user.lastName, _currentUser.user.template);

            if (_currentUser.userGrant != UserGrant.UG_NONE)
            {
                _tcpClient.addUserGrant(_ipAddress, _portNumber, _currentUser.user.firstName, _currentUser.user.lastName, _serialNumber, _currentUser.userGrant);
            }

            if (!String.IsNullOrEmpty(_currentUser.user.BadgeReaderID.Trim()))
            {
                _tcpClient.addUserBadge(_ipAddress, _portNumber, _currentUser.user.firstName, _currentUser.user.lastName, _currentUser.user.BadgeReaderID);
            }

            RefreshListBoxUsers();
        }

        private void buttonStart_Click(object sender, EventArgs e)
        {
            TcpIpClient.RetCode response = _tcpClient.requestScan(_ipAddress, _portNumber, _serialNumber);

            switch (response)
            {
                case TcpIpClient.RetCode.RC_FailedToConnect:
                    UpdateStatusBar("Unable to perform Scan request: Device not reachable.");
                    break;

                default:
                    break;
            }
        }

        private void buttonStop_Click(object sender, EventArgs e)
        {
            _tcpClient.requestStopScan(_ipAddress, _portNumber, _serialNumber);
        }

        /// <summary>
        /// Click on the "Refresh" button (User management tab).
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonRefresh_Click(object sender, EventArgs e)
        {
            RefreshListBoxUsers();
        }

        /// <summary>
        /// Refresh the list of known users by getting an up-to-date list from Device.
        /// Executed by the ThreadPool in order not to block the UI thread.
        /// </summary>
        private void RefreshListBoxUsers()
        {
            ThreadPool.QueueUserWorkItem(new WaitCallback(delegate
            {
                _grantedUsers = null;
                Invoke((MethodInvoker)delegate
                {
                    listBoxUsers.Items.Clear();
                });

                TcpIpClient.RetCode ret = _tcpClient.getUserListWithGrant(_ipAddress, _portNumber, _serialNumber, out _grantedUsers);

                if (_grantedUsers != null)
                {
                    Invoke((MethodInvoker)delegate
                    {
                        listBoxUsers.Items.Clear();
                        listBoxUsers.Items.AddRange(_grantedUsers);
                    });
                }

                else
                {
                    UpdateStatusBar("No user defined");

                }
            }));
        }

        /// <summary>
        /// Event fired when User clicks on an item of the users listbox.
        /// Load the selected granted user's information in the management form.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void listBoxUsers_Click(object sender, EventArgs e)
        {
            if (listBoxUsers.SelectedIndex < 0)
            {
                return;
            }

            _currentUser = (DeviceGrant)listBoxUsers.SelectedItem;

            textBoxFirstName.Text = _currentUser.user.firstName;
            textBoxLastName.Text = _currentUser.user.lastName;
            textBoxBadgeId.Text = _currentUser.user.BadgeReaderID;

            comboBoxGrant.SelectedItem = _currentUser.userGrant;

            buttonAddUser.Visible = false;
            buttonSaveChanges.Visible = true;

            buttonEnroll.Enabled = true;

            textBoxFirstName.ReadOnly = true;
            textBoxLastName.ReadOnly = true;
        }

        /// <summary>
        /// Update grants of the loaded user (by clicking on his name in the list).
        /// The users list is refreshed after the grants have been submitted.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonUpdateGrant_Click(object sender, EventArgs e)
        {
            if (_currentUser == null)
            {
                return;
            }

            UserGrant newGrantValue = (UserGrant)comboBoxGrant.SelectedItem;

            UpdateUserGrant(_currentUser.user.firstName, _currentUser.user.lastName, newGrantValue);
            RefreshListBoxUsers();
        }

        private void UpdateUserGrant(string firstName, string lastName, UserGrant newGrantValue)
        {
            if (newGrantValue != UserGrant.UG_NONE)
            {
                _tcpClient.addUserGrant(_ipAddress, _portNumber, firstName, lastName, _serialNumber, newGrantValue);
            }

            else
            {
                _tcpClient.deleteUserGrant(_ipAddress, _portNumber, firstName, lastName, _serialNumber);
            }
        }

        private void UpdateUserBadge(string firstName, string lastName, string badgeId)
        {
            if (String.IsNullOrEmpty(badgeId.Trim()))
            {
                _tcpClient.deleteUserBadge(_ipAddress, _portNumber, firstName, lastName);
            }

            else
            {
                _tcpClient.addUserBadge(_ipAddress, _portNumber, firstName, lastName, badgeId);
            }
        }

        /// <summary>
        /// App closing event. 
        /// Release fingerprint reader instance (if any instantiated), stop timers and event notifications server (if started).
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainWindow_FormClosing(object sender, FormClosingEventArgs e)
        {
            SetRefreshingTimers(false);

            if (_fpHandler != null)
            {
                _fpHandler.ReleaseFingerprint();
            }

            StopNotificationServer();
        }

        /// <summary>
        /// Remove the currently loaded user from the Device granted users list.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonRemoveUser_Click(object sender, EventArgs e)
        {
            _tcpClient.deleteUserGrant(_ipAddress, _portNumber, _currentUser.user.firstName, _currentUser.user.lastName, _serialNumber);
            _tcpClient.deleteUser(_ipAddress, _portNumber, _currentUser.user.firstName, _currentUser.user.lastName, _serialNumber);

            RefreshListBoxUsers();
            ResetUserForm();
        }

        private void buttonEnroll_Click(object sender, EventArgs e)
        {
            if (listBoxUsers.SelectedIndex < 0)
            {
                MessageBox.Show("Please select a user to enroll fingerprint(s).", "User needed", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }

            _currentUser.user.template = _fpHandler.EnrollUser(null, _currentUser.user.firstName, _currentUser.user.lastName, _currentUser.user.template, false);
        }

        private void buttonReset_Click(object sender, EventArgs e)
        {
            ResetUserForm();
        }

        /// <summary>
        /// Empty/reset the user management form.
        /// </summary>
        private void ResetUserForm()
        {
            textBoxFirstName.ReadOnly = false;
            textBoxLastName.ReadOnly = false;

            textBoxFirstName.Text = String.Empty;
            textBoxLastName.Text = String.Empty;
            textBoxBadgeId.Text = String.Empty;

            buttonAddUser.Visible = true;
            buttonSaveChanges.Visible = false;

            buttonEnroll.Enabled = false;
        }

        /// <summary>
        /// Save changes made on the loaded user.
        /// Refresh users list automatically once all changes have been saved.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonSaveChanges_Click(object sender, EventArgs e)
        {
            string firstName = _currentUser.user.firstName, lastName = _currentUser.user.lastName;

            // update fingerprints templates
            UpdateStatusBar("Saving User...");

            TcpIpClient.RetCode ret =
            _tcpClient.addUserFromTemplate(_ipAddress, _portNumber, firstName, lastName, _currentUser.user.template);

            UpdateStatusBar(ret == TcpIpClient.RetCode.RC_Succeed ? "Changes successfully saved." : "An error occurred while saving changes.");

            // update grants & badge id.
            if ((UserGrant)comboBoxGrant.SelectedItem != _currentUser.userGrant)
            {
                UpdateUserGrant(firstName, lastName, (UserGrant)comboBoxGrant.SelectedItem);
            }

            string inputBadgeId = textBoxBadgeId.Text.Trim();
            if (inputBadgeId != _currentUser.user.BadgeReaderID)
            {
                UpdateUserBadge(firstName, lastName, inputBadgeId);
            }

            RefreshListBoxUsers();
        }
       

        /// <summary>
        /// Refreshing current temperature Timer tick.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void timerRefreshCurrentTemperature_Tick(object sender, EventArgs e)
        {
            ThreadPool.QueueUserWorkItem(new WaitCallback(delegate
            {
                tempInfo currentTempInfo = new tempInfo();

                _tcpClient.getFridgeCurrentTemp(_ipAddress, _portNumber, _serialNumber, out currentTempInfo);

                if (currentTempInfo != null)
                {
                    // Called on UI thread.
                    Invoke((MethodInvoker)delegate { textBoxCurrentTemperature.Text = String.Format("{0:0.0}", currentTempInfo.lastTempValue); });
                }
            }));
        }

        /// <summary>
        /// Click on the "Refresh" button of Temperature tab.
        /// Reload the measures made during last 12 hours and paint the chart again.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonRefreshChart_Click(object sender, EventArgs e)
        {
            _chartTemperature.Series[SERIES_T_NAME].Points.Clear();
            buttonRefreshChart.Enabled = false;

            ThreadPool.QueueUserWorkItem(new WaitCallback(delegate
            {
                int amountOfHourBack = 4;
                DateTime startDate = DateTime.Now.AddHours((amountOfHourBack + 1) * -1);

                tempInfo[] lastTempInfoGroup;
                _tcpClient.getFridgeTempFromDate(_ipAddress, _portNumber, _serialNumber, startDate, out lastTempInfoGroup);

                if (lastTempInfoGroup == null || lastTempInfoGroup.Length == 0)
                {
                    return;
                }

                // points displayed in the Chart object. string (X) is the date + hour the temperature was measured. double (Y) is the temperature value.
                Dictionary<string, double?> newPoints = new Dictionary<string, double?>();

                // determine if there are temperature measures in the tempBottle array (depends on Fridge type and manufacturer).
                bool useBottleTemperature = lastTempInfoGroup[lastTempInfoGroup.Length - 1].tempBottle.Count > 0;

                DateTime previousTempInfoDate = lastTempInfoGroup[0].CreationDate.AddHours(-1);

                for (int i = 0; i < lastTempInfoGroup.Length; ++i)
                {
                    tempInfo currentTempInfo = lastTempInfoGroup[i];
                    DateTime lotDate = currentTempInfo.CreationDate;

                    // if one (or more) hour is missing from measures list, let's add empty points in the chart.
                    while ((lotDate - previousTempInfoDate).TotalMinutes > 61)
                    {
                        previousTempInfoDate = previousTempInfoDate.AddHours(1);

                        for (int j = 0; j < 60; ++j)
                        {
                            string pointDate = previousTempInfoDate.AddMinutes(j).ToString("dd/MM - HH:mm");
                            newPoints[pointDate] = null;
                        }
                    }

                    // first temperature minute time (e.g "29" if it was measured at 5:29pm)
                    int minuteStart = lotDate.Minute;

                    // select the Dictionary to be used in order to get temperature values. If there are not measures in tempBottle array, we must use tempArray.
                    Dictionary<int, double> minuteToTemperature = ((useBottleTemperature) ? currentTempInfo.tempBottle : currentTempInfo.tempArray);

                    // for all points missing, from hh:00 to hh:minuteStart, add empty points in the chart.
                    for (int j = 0; j < minuteStart; ++j)
                    {
                        string pointDate = lotDate.AddMinutes(-minuteStart + j).ToString("dd/MM - HH:mm");
                        newPoints[pointDate] = null;
                    }

                    // for all points from hh:minuteStart to the end of the hour, add the point value (if any) or an empty point.
                    for (int j = minuteStart; j < 60; ++j)
                    {
                        string pointDate = lotDate.AddMinutes(j - minuteStart).ToString("dd/MM - HH:mm");
                        newPoints[pointDate] = minuteToTemperature.ContainsKey(j) ? minuteToTemperature[j] : (double?) null;
                    }

                    // update the last processed datetime.
                    previousTempInfoDate = lotDate;
                }

                // draw all the measures (executed on UI thread).
                Invoke((MethodInvoker)delegate
                {
                    foreach (KeyValuePair<string, double?> newPoint in newPoints)
                    {
                        _chartTemperature.Series[SERIES_T_NAME].Points.AddXY(newPoint.Key, newPoint.Value);
                    }

                    buttonRefreshChart.Enabled = true;
                });
            }));
        }

        private void timerRefreshConnectionStatus_Tick(object sender, EventArgs e)
        {
            ThreadPool.QueueUserWorkItem(new WaitCallback(delegate
            {
                if (_tcpClient.pingDevice(_ipAddress, _portNumber, _serialNumber) != TcpIpClient.RetCode.RC_Succeed)
                {
                    UpdateStatusBar("Disconnected from device.");
                    Invoke((MethodInvoker)delegate
                    {
                        tabControl.Enabled = false;
                        pictureBoxConnectionStatus.Image = RED_ICON;
                    });
                    return;
                }

                Invoke((MethodInvoker)delegate
                {
                    tabControl.Enabled = true;
                    pictureBoxConnectionStatus.Image = GREEN_ICON;
                });
            }));
        }

        private void tabControl_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (tabControl.TabPages[1].Visible)
                RefreshListBoxUsers();
        }

        private void listBoxTag_SelectedIndexChanged(object sender, EventArgs e)
        {
            buttonLight.Enabled = listBoxTag.SelectedItems.Count > 0;
        }

        private void btLight_Click(object sender, EventArgs e)
        {
            //List of tag to light
            List<string> tagsListToLight = listBoxTag.SelectedItems.Cast<string>().ToList();
            int nbTagToLight = tagsListToLight.Count; // initial number of tags

            TcpIpClient.RetCode ret;
            //Request Lighting process.
            //MANDATORY - A new scan has to be perform after each boot of the device to have the light working.
            //All tag find and light a removed from the list.
            //Tag not find stay in the list at end of the process.
            ret = _tcpClient.RequestStartLighting(_ipAddress, _portNumber, tagsListToLight);
            if (ret != TcpIpClient.RetCode.RC_Succeed)
            {
                MessageBox.Show("An unexpected error occurred during communication with device : " + ret.ToString());
                _tcpClient.RequestStopLighting(_ipAddress, _portNumber);
                return;
            }
            StringBuilder resultMessage = new StringBuilder(String.Format("{0} tags to find : {1} have been found.", nbTagToLight,
                nbTagToLight - tagsListToLight.Count));

            if (tagsListToLight.Count > 0) // some tag UIDs are still in the list : they've not been found
            {
                resultMessage.AppendLine("Missing tags ID :");

                foreach (string missingTag in tagsListToLight)
                    resultMessage.AppendLine(missingTag);
            }

            // this is not mandatory , just for info.
            MessageBox.Show(resultMessage.ToString());

            // Light is stopped when start started if use with access badge so this function is not mandatory if user open door.
            //if user not open the fridge it stop the light process
            _tcpClient.RequestStopLighting(_ipAddress, _portNumber);

            
            /*TcpIpClient.RetCode response = _tcpClient.requestScan(_ipAddress, _portNumber, _serialNumber);

            switch (response)
            {
                case TcpIpClient.RetCode.RC_FailedToConnect:
                    UpdateStatusBar("Unable to perform Scan request: Device not reachable.");
                    break;

                default:
                    break;
            }
            string strStatus = string.Empty;
            do
            {
                
                Thread.Sleep(1000);
                _tcpClient.getStatus(_ipAddress, _portNumber, _serialNumber, out strStatus);
            } while (strStatus != "DS_Ready");
            
            
            
                //List of tag to light
                tagsListToLight = listBoxTag.SelectedItems.Cast<string>().ToList();
                nbTagToLight = tagsListToLight.Count; // initial number of tags

                
                //Request Lighting process.
                //MANDATORY - A new scan has to be perform after each boot of the device to have the light working.
                //All tag find and light a removed from the list.
                //Tag not find stay in the list at end of the process.
                ret = _tcpClient.RequestStartLighting(_ipAddress, _portNumber, tagsListToLight);
                if (ret != TcpIpClient.RetCode.RC_Succeed)
                {
                    MessageBox.Show("An unexpected error occurred during communication with device : " + ret.ToString());
                    _tcpClient.RequestStopLighting(_ipAddress, _portNumber);
                    return;
                }
                resultMessage = new StringBuilder(String.Format("{0} tags to find : {1} have been found.", nbTagToLight,
                    nbTagToLight - tagsListToLight.Count));

                if (tagsListToLight.Count > 0) // some tag UIDs are still in the list : they've not been found
                {
                    resultMessage.AppendLine("Missing tags ID :");

                    foreach (string missingTag in tagsListToLight)
                        resultMessage.AppendLine(missingTag);
                }

                // this is not mandatory , just for info.
                MessageBox.Show(resultMessage.ToString());

                // Light is stopped when start started if use with access badge so this function is not mandatory if user open door.
                //if user not open the fridge it stop the light process
                _tcpClient.RequestStopLighting(_ipAddress, _portNumber);*/

        }

    }
}
