﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace SampleFridge
{
    public static class Tool
    {
        /// <summary>
        /// Allow verification of IP Address & Port Number format.
        /// </summary>
        /// <param name="givenIp">Variable used to store IP Address.</param>
        /// <param name="portNumber">Variable used to store Port Number.</param>
        /// <param name="ipInput">User input for IP Address.</param>
        /// <param name="portInput">User input for Port Number.</param>
        /// <returns>True if parameters are valid, False otherwise.</returns>
        public static bool TryParseIpPort(out string givenIp, out int portNumber, string ipInput, string portInput)
        {
            givenIp = ipInput.Trim();
            portNumber = 0;

            string givenPort = portInput.Trim();

            if (String.IsNullOrEmpty(givenIp) || String.IsNullOrEmpty(givenPort))
            {
                return false;
            }

            IPAddress ipAddress;

            if (!IPAddress.TryParse(givenIp, out ipAddress)) // check if IP address is valid
            {
                return false;
            }

            if (ipAddress.AddressFamily != System.Net.Sockets.AddressFamily.InterNetwork) // check if IP is IPv4
            {
                return false;
            }

            if (!int.TryParse(givenPort, out portNumber))
            {
                return false;
            }

            return true;
        }
    }
}
